from __future__ import annotations

import logging
from enum import Enum
from typing import (
    Any,
    Dict,
    Final,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

from scrapli.exceptions import ScrapliTransportPluginError
from scrapli.response import Response

from .base import BaseWattBox, Commands, Outlet, _async_create_wattbox, _create_wattbox
from .driver.async_driver import WattBoxAsyncDriver
from .driver.sync_driver import WattBoxDriver

logger = logging.getLogger("pywattbox.ip")


class REQUEST_MESSAGES(Enum):
    FIRMWARE = "?Firmware"
    HOSTNAME = "?Hostname"
    SERVICE_TAG = "?ServiceTag"
    MODEL = "?Model"
    OUTLET_COUNT = "?OutletCount"
    OUTLET_STATUS = "?OutletStatus"
    OUTLET_POWER_STATUS = "?OutletPowerStatus={outlet}"
    POWER_STATUS = "?PowerStatus"
    AUTO_REBOOT = "?AutoReboot"
    OUTLET_NAME = "?OutletName"
    UPS_STATUS = "?UPSStatus"
    UPS_CONNECTION = "?UPSConnection"


class CONTROL_MESSAGES(Enum):
    OUTLET_NAME_SET = "!OutletNameSet={outlet},{name}"
    OUTLET_NAME_SET_ALL = "!OutletNameSetAll={names}"  # Comma separated list of names
    OUTLET_SET = "!OutletSet={outlet},{action},{delay}"  # Optional delay
    OUTLET_POWER_ON_DELAY_SET = "!OutletPowerOnDelaySet={outlet},{delay}"
    OUTLET_MODE_SET = "!OutletModeSet={outlet},{mode}"
    OUTLET_REBOOT_SET = "!OutletRebootSet={ops}"  # Comma separate list of ops
    AUTO_REBOOT = "!AutoReboot={state}"
    AUTO_REBOOT_TIMEOUT_SET = (
        "!AutoRebootTimeoutSet={timeout},{count},{ping_delay},{reboot_attempts}"
    )
    FIRMWARE_UPDATE = "!FirmwareUpdate={url}"
    REBOOT = "!Reboot"
    ACCOUNT_SET = "!AccountSet={user},{password}"
    NETWORK_SET = (
        "!NetworkSet={host},{ip},{subnet},{gateway},{dns1},{dns2}"  # DNS 2 is optional
    )
    SCHEDULE_ADD = "!ScheduleAdd={schedule}"
    HOST_ADD = "!HostAdd={name},{ip},{outlets}"
    SET_TELNET = "!SetTelnet={mode}"
    WEB_SERVER_SET = "!WebServerSet={mode}"
    SET_SDDP = "!SetSDDP={mode}"


INITIAL_REQUESTS: Final[Tuple[REQUEST_MESSAGES, ...]] = (
    REQUEST_MESSAGES.MODEL,
    REQUEST_MESSAGES.FIRMWARE,
    REQUEST_MESSAGES.UPS_CONNECTION,
    REQUEST_MESSAGES.HOSTNAME,
    REQUEST_MESSAGES.SERVICE_TAG,
    REQUEST_MESSAGES.OUTLET_COUNT,
)


class InitialResponses(NamedTuple):
    hardware_version: Response
    firmware_version: Response
    has_ups: Response
    hostname: Response
    serial_number: Response
    number_outlets: Response


UPDATE_BASE_REQUESTS: Final[Tuple[REQUEST_MESSAGES, ...]] = (
    REQUEST_MESSAGES.AUTO_REBOOT,
    REQUEST_MESSAGES.POWER_STATUS,
    REQUEST_MESSAGES.OUTLET_NAME,
    REQUEST_MESSAGES.OUTLET_STATUS,
)


class UpdateBaseResponses(NamedTuple):
    auto_reboot: Response
    power_status: Response
    outlet_name: Response
    outlet_status: Response


_Responses = TypeVar("_Responses", bound=Union[InitialResponses, UpdateBaseResponses])


class DriverUnavailableError(Exception):
    pass


class IpWattBox(BaseWattBox):
    def __init__(
        self,
        host: str,
        user: str,
        password: str,
        port: int = 22,
        transport: Optional[str] = None,
    ) -> None:
        super().__init__(host, user, password, port)

        self.battery_test = None
        self.cloud_status = None
        self.outlet_power_status: bool = False

        conninfo: Dict[str, Any] = {
            "host": host,
            "auth_username": user,
            "auth_password": password,
            "port": port,
        }
        if transport is None:
            if port == 22:
                transport = "ssh"
            elif port == 23:
                transport = "telnet"
            else:
                raise ValueError("Non Standard Port, Transport must be set.")

        try:
            self.driver: Optional[WattBoxDriver] = WattBoxDriver(
                **conninfo,
                transport="ssh2" if transport == "ssh" else "telnet",
            )
        except ScrapliTransportPluginError:
            self.driver = None
        try:
            self.async_driver: Optional[WattBoxAsyncDriver] = WattBoxAsyncDriver(
                **conninfo,
                transport="asyncssh" if transport == "ssh" else "asynctelnet",
            )
        except ScrapliTransportPluginError:
            self.async_driver = None

    def send_requests(
        self, requests: Iterable[Union[REQUEST_MESSAGES, str]]
    ) -> List[Response]:
        if not self.driver:
            raise DriverUnavailableError
        responses: List[Response] = []
        for request in requests:
            responses.append(
                self.driver._send_command(
                    request.value if isinstance(request, REQUEST_MESSAGES) else request
                )
            )
        return responses

    async def async_send_requests(
        self, requests: Iterable[Union[REQUEST_MESSAGES, str]]
    ) -> List[Response]:
        if not self.async_driver:
            raise DriverUnavailableError
        responses: List[Response] = []
        for request in requests:
            responses.append(
                await self.async_driver._send_command(
                    request.value if isinstance(request, REQUEST_MESSAGES) else request
                )
            )
        return responses

    def parse_initial(self, responses: InitialResponses) -> None:
        logger.debug("Parse Initial")
        # TODO: Add if failed logic?
        self.hardware_version = responses.hardware_version.result
        if "150" not in self.hardware_version and "250" not in self.hardware_version:
            self.outlet_power_status = True
        self.firmware_version = responses.firmware_version.result
        self.has_ups = responses.has_ups.result == "1"
        self.hostname = responses.hostname.result
        self.serial_number = responses.serial_number.result
        self.number_outlets = (
            int(count) if (count := responses.number_outlets.result) else 0
        )
        # The index for outlet within WattBox starts at 1.
        self.outlets = {i: Outlet(i, self) for i in range(1, self.number_outlets + 1)}

    def get_initial(self) -> None:
        logger.debug("Get Initial")
        responses = InitialResponses(*self.send_requests(INITIAL_REQUESTS))
        self.parse_initial(responses)

    async def async_get_initial(self) -> None:
        logger.debug("Async Get Initial")
        responses = InitialResponses(
            *(await self.async_send_requests(INITIAL_REQUESTS))
        )
        logger.debug("Responses: %s", responses)
        self.parse_initial(responses)

    def parse_update_base(self, responses: UpdateBaseResponses) -> None:
        logger.debug("Parse Update Base")
        # auto reboot
        self.auto_reboot = responses.auto_reboot.result == "1"
        # power status
        power_status = responses.power_status.result.split(",")
        self.current_value = float(power_status[0])
        self.power_value = float(power_status[1])
        self.voltage_value = float(power_status[2])
        # The light is green and shows as green in the web UI, but strangely
        # the value is "0" in this API call.
        self.safe_voltage_status = power_status[3] == "0"
        # outlet_name
        for i, s in enumerate(responses.outlet_name.result.split(","), start=1):
            self.outlets[i].name = s.lstrip("{").rstrip("}")
        # outlet_status
        for i, s in enumerate(responses.outlet_status.result.split(","), start=1):
            self.outlets[i].status = s == "1"

    def parse_ups_status(self, response: Response) -> None:
        logger.debug("Parse UPS Status")
        ups_status = response.result.split(",")
        self.battery_charge = int(ups_status[0])
        self.battery_load = int(ups_status[1])
        self.battery_health = ups_status[2] == "Good"
        self.power_lost = ups_status[3] == "True"
        self.est_run_time = int(ups_status[4])
        self.audible_alarm = ups_status[5] == "True"
        self.mute = ups_status[6] == "True"

    def parse_outlet_power_statuses(self, responses: Iterable[Response]) -> None:
        logger.debug("Parse Outlet Statuses")
        for response in responses:
            index, power, current, voltage = response.result.split(",")
            # The index in the python list is off by 1 from the WattBox index.
            outlet = self.outlets[int(index)]
            outlet.power_value = float(power)
            outlet.current_value = float(current)
            outlet.voltage_value = float(voltage)

    @property
    def update_requests(self) -> Tuple[Union[REQUEST_MESSAGES, str], ...]:
        return (
            *UPDATE_BASE_REQUESTS,
            REQUEST_MESSAGES.UPS_STATUS,
            *(
                (
                    REQUEST_MESSAGES.OUTLET_POWER_STATUS.value.format(
                        outlet=(outlet.index)
                    )
                    for outlet in self.outlets.values()
                )
                if self.outlet_power_status
                else ()
            ),
        )

    def update(self) -> None:
        logger.debug("Update")
        responses = self.send_requests(self.update_requests)
        self.parse_update_base(UpdateBaseResponses(*responses[0:4]))
        self.parse_ups_status(responses[4])
        if self.outlet_power_status:
            self.parse_outlet_power_statuses(responses[5:])

    async def async_update(self) -> None:
        logger.debug("Async Update")
        responses = await self.async_send_requests(self.update_requests)
        self.parse_update_base(UpdateBaseResponses(*responses[0:4]))
        self.parse_ups_status(responses[4])
        if self.outlet_power_status:
            self.parse_outlet_power_statuses(responses[5:])

    def send_command(self, outlet: int, command: Commands) -> None:
        logger.debug("Send Command")
        if not self.driver:
            raise DriverUnavailableError
        self.driver._send_command(
            CONTROL_MESSAGES.OUTLET_SET.value.format(
                outlet=outlet, action=command.name, delay=0
            )
        )
        self.update()

    async def async_send_command(self, outlet: int, command: Commands) -> None:
        logger.debug("Async Send Command")
        if not self.async_driver:
            raise DriverUnavailableError
        await self.async_driver._send_command(
            CONTROL_MESSAGES.OUTLET_SET.value.format(
                outlet=outlet, action=command.name, delay=0
            )
        )
        await self.async_update()

    # String Representation
    def __str__(self) -> str:
        return f"{self.hostname} ({self.host}): {self.hardware_version}"


def create_ip_wattbox(host: str, user: str, password: str, port: int = 22) -> IpWattBox:
    return _create_wattbox(
        IpWattBox, host=host, user=user, password=password, port=port
    )


async def async_create_ip_wattbox(
    host: str, user: str, password: str, port: int = 22
) -> IpWattBox:
    return await _async_create_wattbox(
        IpWattBox, host=host, user=user, password=password, port=port
    )



from __future__ import annotations

import logging
from io import BytesIO
from typing import Any, Callable, Dict, Optional, Union

from scrapli.decorators import timeout_modifier
from scrapli.driver import Driver
from scrapli.exceptions import ScrapliConnectionNotOpened
from scrapli.response import Response

from . import PROMPTS

logger = logging.getLogger("pywattbox.sync_driver")


def on_open(driver: WattBoxDriver) -> None:
    if driver.transport_name not in ("telnet", "asynctelnet"):
        driver.channel._read_until_prompt()


def on_close(driver: WattBoxDriver) -> None:
    try:
        driver.channel.write("!Exit")
        driver.channel.send_return()
    except ScrapliConnectionNotOpened:
        pass


class WattBoxDriver(Driver):
    def __init__(
        self,
        host: str,
        port: Optional[int] = 22,
        auth_username: str = "",
        auth_password: str = "",
        auth_private_key: str = "",
        auth_private_key_passphrase: str = "",
        auth_strict_key: bool = False,
        auth_bypass: bool = False,
        timeout_socket: float = 5.0,
        timeout_transport: float = 5.0,
        timeout_ops: float = 5.0,
        comms_prompt_pattern: str = PROMPTS,
        comms_return_char: str = "\n",
        ssh_config_file: Union[str, bool] = False,
        ssh_known_hosts_file: Union[str, bool] = False,
        on_init: Optional[Callable[..., Any]] = None,
        on_open: Optional[Callable[..., Any]] = on_open,
        on_close: Optional[Callable[..., Any]] = on_close,
        transport: str = "ssh2",
        transport_options: Optional[Dict[str, Any]] = None,
        channel_log: Union[str, bool, BytesIO] = False,
        channel_log_mode: str = "write",
        channel_lock: bool = True,
        logging_uid: str = "",
    ) -> None:
        super().__init__(
            host=host,
            port=port,
            auth_username=auth_username,
            auth_password=auth_password,
            auth_private_key=auth_private_key,
            auth_private_key_passphrase=auth_private_key_passphrase,
            auth_strict_key=auth_strict_key,
            auth_bypass=auth_bypass,
            timeout_socket=timeout_socket,
            timeout_transport=timeout_transport,
            timeout_ops=timeout_ops,
            comms_prompt_pattern=comms_prompt_pattern,
            comms_return_char=comms_return_char,
            ssh_config_file=ssh_config_file,
            ssh_known_hosts_file=ssh_known_hosts_file,
            on_init=on_init,
            on_open=on_open,
            on_close=on_close,
            transport=transport,
            transport_options=transport_options,
            channel_log=channel_log,
            channel_log_mode=channel_log_mode,
            channel_lock=channel_lock,
            logging_uid=logging_uid,
        )

    def _open(self, force: bool = False) -> None:
        if force or not self.transport.isalive():
            self.open()

    @timeout_modifier
    def _send_command(
        self,
        command: str,
    ) -> Response:
        """Send a command.

        Based on:
            scrapli.driver.generic.sync_driver.GenericDriver: send_command and _send_command
            scrapli.channel.sync_channel.Channel: send_input

        Args:
            command: string to send to device in privilege exec mode
            failed_when_contains: string or list of strings indicating failure if found in response

        Returns:
            Response: Scrapli Response object
        """
        self._open()

        response = Response(
            host=self._base_transport_args.host,
            channel_input=command,
            failed_when_contains="#Error",
        )

        # Normally handled in the channel `send_input`, but WattBox is special and doesn't work
        # with that function. Pulled it all into the Driver for simplicity.
        with self.channel._channel_lock():
            self.channel.write(command)
            self.channel.send_return()
            raw_response = self.channel._read_until_prompt()

            logger.debug("raw_response: %s", raw_response)
            split_response = raw_response.strip().splitlines()
            logger.debug("split_response: %s", split_response)
            if (
                self.transport not in ("telnet", "asynctelnet")
                and len(split_response) < 2
            ):
                logger.error("Not enough lines: %s. Getting more", len(split_response))
                raw_response += self.channel._read_until_prompt()
                logger.debug("raw_response: %s", raw_response)
                split_response = raw_response.strip().splitlines()
                logger.debug("split_response: %s", split_response)

        if (
            self.transport not in ("telnet", "asynctelnet")
            and split_response[0] != command.encode()
        ):
            logger.error("Doesn't match command: %s - %s", command, split_response[0])

        if command.startswith("?"):
            if not split_response[-1].startswith(command.encode()):
                logger.error(
                    "Expected response to start with: %s, Got %s",
                    command,
                    split_response[-1],
                )
            processed_response = split_response[1].split(b"=")[-1]
        else:
            processed_response = split_response[-1]

        logger.debug("processed_response: %s", processed_response)
        response.record_response(processed_response)
        response.raw_result = raw_response
        return response



from typing import Final

PROMPTS: Final[str] = (
    r"^"  # Start of the string
    r"(.*Successfully Logged In!)|"  # After Login
    r"(\?\w+=\S+)|"  # Response to `?` request message
    r"(OK)|"  # Response to `!` control message
    r"(#Error)"  # Error Message
    r"\n$"  # Newline / End of String
)



from __future__ import annotations

import logging
from abc import ABC, abstractmethod
from enum import IntEnum
from typing import Dict, Optional, Type, TypeVar

logger = logging.getLogger("pywattbox")


class Commands(IntEnum):
    """Commands Enum for Convenience.

    HTTP API uses the values.
    Integration Protocol uses the names.
    """

    OFF = 0
    ON = 1
    RESET = 3
    # Only used for HTTP
    AUTO_REBOOT_ON = 4
    AUTO_REBOOT_OFF = 5
    # Only used for Integration Protocol
    TOGGLE = 6


class BaseWattBox(ABC):
    """Base WattBox that defines the"""

    def __init__(self, host: str, user: str, password: str, port: int) -> None:
        self.host: str = host
        self.port: Optional[int] = port
        self.user: str = user
        self.password: str = password

        # Info, set once
        self.hardware_version: Optional[str] = None
        self.firmware_version: Optional[str] = None
        self.has_ups: bool = False
        self.hostname: str = ""
        self.number_outlets: int = 0
        self.serial_number: str = ""

        # Status values
        self.audible_alarm: bool = False
        self.auto_reboot: bool = False
        self.cloud_status: Optional[bool] = False
        self.mute: bool = False
        self.power_lost: bool = False

        # Power values
        self.current_value: float = 0.0  # In Amps
        self.power_value: float = 0.0  # In watts
        self.safe_voltage_status: bool = True
        self.voltage_value: float = 0.0  # In volts

        # Battery values
        self.battery_charge: int = 0  # In percent
        self.battery_health: bool = False
        self.battery_load: int = 0  # In percent
        self.battery_test: Optional[bool] = False
        self.est_run_time: int = 0  # In minutes

        # Outlets list
        self.outlets: Dict[int, Outlet] = {}
        self.master_outlet: Optional[Outlet] = None

    @abstractmethod
    def get_initial(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def async_get_initial(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def update(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def async_update(self) -> None:
        raise NotImplementedError()

    @abstractmethod
    def send_command(self, outlet: int, command: Commands) -> None:
        raise NotImplementedError()

    @abstractmethod
    async def async_send_command(self, outlet: int, command: Commands) -> None:
        raise NotImplementedError()


_T_WattBox = TypeVar("_T_WattBox", bound=BaseWattBox)


def _create_wattbox(
    type_: Type[_T_WattBox], host: str, user: str, password: str, port: int
) -> _T_WattBox:
    wattbox = type_(host=host, user=user, password=password, port=port)
    wattbox.get_initial()
    wattbox.update()
    return wattbox


async def _async_create_wattbox(
    type_: Type[_T_WattBox], host: str, user: str, password: str, port: int
) -> _T_WattBox:
    wattbox = type_(host=host, user=user, password=password, port=port)
    await wattbox.async_get_initial()
    await wattbox.async_update()
    return wattbox


class Outlet:
    def __init__(self, index: int, wattbox: BaseWattBox) -> None:
        self.index: int = index
        self.method: Optional[bool] = None
        self.name: Optional[str] = ""
        self.status: Optional[bool] = None
        # Power values
        self.current_value: Optional[float] = None  # In Amps
        self.power_value: Optional[float] = None  # In watts
        self.voltage_value: Optional[float] = None  # In volts
        # The WattBox
        self.wattbox: BaseWattBox = wattbox

    def turn_on(self) -> None:
        self.wattbox.send_command(self.index, Commands.ON)

    async def async_turn_on(self) -> None:
        await self.wattbox.async_send_command(self.index, Commands.ON)

    def turn_off(self) -> None:
        self.wattbox.send_command(self.index, Commands.OFF)

    async def async_turn_off(self) -> None:
        await self.wattbox.async_send_command(self.index, Commands.OFF)

    def reset(self) -> None:
        self.wattbox.send_command(self.index, Commands.RESET)

    async def async_reset(self) -> None:
        await self.wattbox.async_send_command(self.index, Commands.RESET)

    def __str__(self) -> str:
        return f"{self.name} ({self.index}): {self.status}"
